# -*- coding: utf-8 -*-
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Connector between api.py and components."""
import json
import os
from threading import Thread
from typing import Any, Callable, Dict, List, Optional

from werkzeug.wrappers import Response as WebResponse

from neural_insights.components.diagnosis.diagnosis import Diagnosis
from neural_insights.components.diagnosis.factory import DiagnosisFactory
from neural_insights.components.diagnosis.op_details import OpDetails
from neural_insights.components.graph.graph import Graph
from neural_insights.components.graph.graph_reader import GraphReader
from neural_insights.components.workload_manager.workload_manager import WorkloadManager
from neural_insights.utils.exceptions import ClientErrorException
from neural_insights.utils.json_serializer import JsonSerializer
from neural_insights.web.communication import Request, Response, create_simple_response
from neural_insights.web.exceptions import ServiceNotFoundException
from neural_insights.web.service.request_data_processor import RequestDataProcessor


class RoutingDefinition:
    """Abstract routing definition."""

    def __init__(self, callback: Callable[[Dict[str, Any]], Any]) -> None:
        """Initialize object."""
        self.callback = callback


class RealtimeRoutingDefinition(RoutingDefinition):
    """Routing executed in realtime."""

    pass


class DeferredRoutingDefinition(RoutingDefinition):
    """Routing executed in separate thread."""

    pass


class Router:
    """Connector between api.py and components."""

    def __init__(self) -> None:
        """Initialize object."""
        self.routes: Dict[str, RoutingDefinition] = {
            "workloads": RealtimeRoutingDefinition(get_workloads_list),
            "workloads/delete": RealtimeRoutingDefinition(delete_workload),
            "profiling": RealtimeRoutingDefinition(get_profiling_details),
            "model/graph": RealtimeRoutingDefinition(get_model_graph),
            "model/summary": RealtimeRoutingDefinition(get_model_summary),
            "model/graph/highlight_pattern": RealtimeRoutingDefinition(find_pattern_in_graph),
            "diagnosis/op_list": RealtimeRoutingDefinition(get_op_list),
            "diagnosis/op_details": RealtimeRoutingDefinition(get_op_details),
            "diagnosis/histogram": RealtimeRoutingDefinition(get_histogram),
            "profiling/result": RealtimeRoutingDefinition(get_profiling_details),
        }

    def handle(self, request: Request) -> Response:
        """Run operation on requested component and return result."""
        routing_definition = self.routes.get(request.operation)
        if routing_definition is None:
            raise ServiceNotFoundException(f"Unable to find {request.operation}")

        data = self._process_routing_definition(routing_definition, request.data)
        if isinstance(data, WebResponse):
            return data

        serialized_data = JsonSerializer.serialize_item(data)

        return create_simple_response(serialized_data)

    def _process_routing_definition(
        self,
        routing_definition: RoutingDefinition,
        data: dict,
    ) -> Any:
        """Process routing definition."""
        if isinstance(routing_definition, RealtimeRoutingDefinition):
            return routing_definition.callback(data)
        if isinstance(routing_definition, DeferredRoutingDefinition):
            self._validate_deffered_routing_data(data)
            t = Thread(target=routing_definition.callback, args=(data,))
            t.daemon = True
            t.start()
            return {"exit_code": 102, "message": "processing"}
        raise ValueError(
            f"Unsupported RoutingDefinition type: {routing_definition.__class__.__name__}",
        )

    @staticmethod
    def _validate_deffered_routing_data(data: dict) -> None:
        """Validate input data for Deferred Routing and raises in case of issues."""
        request_id = str(data.get("request_id", ""))
        if not request_id:
            raise ClientErrorException("Missing request id.")


def get_model_graph(data: Dict[str, Any]) -> Graph:
    """Get model graph."""
    graph_reader = GraphReader()
    return graph_reader.read(
        model_path=RequestDataProcessor.get_string_value(data, "path"),
        expanded_groups=data.get("group", []),
    )


def get_model_summary(data: Dict[str, Any]) -> Dict:
    """Get model graph."""
    workload_id: Optional[str] = data.get("workload_id", None)
    workload = WorkloadManager().get_workload(workload_id)

    if workload.model_summary_file is None:
        raise Exception("Model summary not found.")
    with open(workload.model_summary_file, "r") as summary_file:
        model_summary = "\n".join(summary_file.readlines())
    return {
        "summary": model_summary,
    }


def find_pattern_in_graph(data: Dict[str, Any]) -> dict:
    """Find OP pattern in graph for diagnosis tab."""
    graph_reader = GraphReader()

    model_path = RequestDataProcessor.get_string_value(data, "path")
    op_name = data.get("op_name", None)
    pattern = data.get("pattern", None)
    if any(param is None for param in [model_path, op_name, pattern]):
        raise ClientErrorException(
            "Missing parameters. Required parameters are: path, op_name and pattern.",
        )
    model_graph, expanded_groups = graph_reader.find_pattern_in_graph(
        model_path=model_path,
        op_name=op_name,
        pattern=pattern,
    )
    return {"graph": model_graph.serialize(), "groups": expanded_groups}


def get_workloads_list(data: Dict[str, Any]) -> dict:
    """Get workloads list."""
    workload_manager = WorkloadManager()

    serialized_workloads = [workload.serialize() for workload in workload_manager.workloads]
    return {
        "workloads": serialized_workloads,
    }


def delete_workload(data: Dict[str, Any]) -> dict:
    """Remove workload from workloads list."""
    workload_id: Optional[str] = data.get("workload_id", None)
    if workload_id is None:
        raise ClientErrorException("Could not find workload ID.")

    removed_id = WorkloadManager().remove_workload(workload_id)

    return {
        "workload_id": removed_id,
    }


def get_diagnosis(workload_id: str) -> Diagnosis:
    """Get diagnosis object for specified workload."""
    workload = WorkloadManager().get_workload(workload_id)
    diagnosis = DiagnosisFactory.get_diagnosis(workload)
    return diagnosis


def get_op_list(data: Dict[str, Any]) -> List[dict]:
    """Get OP list for model."""
    workload_id: Optional[str] = data.get("workload_id", None)
    if workload_id is None:
        raise ClientErrorException("Could not find workload ID.")
    diagnosis = get_diagnosis(workload_id)

    return diagnosis.get_op_list()


def get_op_details(data: Dict[str, Any]) -> dict:
    """Get OP details for specific OP in model."""
    try:
        workload_id: str = str(data.get("workload_id", None))
        op_name: str = str(data.get("op_name", None))
    except ValueError:
        raise ClientErrorException("Incorrect parameter values.")
    except TypeError:
        raise ClientErrorException("Could not find all required parameters.")

    diagnosis = get_diagnosis(workload_id)

    op_details: Optional[OpDetails] = diagnosis.get_op_details(op_name)
    if op_details is None:
        return {}
    return op_details.serialize()


def get_histogram(data: Dict[str, Any]) -> list:
    """Get histogram of specific tensor in model."""
    try:
        workload_id: str = str(data.get("workload_id", None))
        op_name: str = str(data.get("op_name", None))
        histogram_type: str = str(data.get("type", None))
    except ValueError:
        raise ClientErrorException("Incorrect parameter values.")
    except TypeError:
        raise ClientErrorException("Could not find all required parameters.")

    diagnosis = get_diagnosis(workload_id)

    histogram_type_map = {
        "weights": "weight",
        "activation": "activation",
    }

    parsed_histogram_type: Optional[str] = histogram_type_map.get(histogram_type, None)
    if parsed_histogram_type is None:
        raise ClientErrorException(
            f"Histogram type not supported. " f"Use one of following: {histogram_type_map.keys()}",
        )

    histogram_data = diagnosis.get_histogram_data(op_name, parsed_histogram_type)
    return histogram_data


def get_profiling_details(data: Dict[str, Any]) -> List[dict]:
    """Get profiling result."""
    workload_id: Optional[str] = data.get("workload_id", None)
    if workload_id is None:
        raise ClientErrorException("Could not find workload ID.")
    workload = WorkloadManager().get_workload(workload_id)
    profiling_data_path = os.path.join(
        workload.workload_location,
        "profiling_data.json",
    )
    with open(profiling_data_path, "r") as json_file:
        profiling_data = json.load(json_file)
    return profiling_data



# -*- coding: utf-8 -*-
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test Collapser."""

import unittest
from typing import List

from neural_insights.components.graph.collapser import Collapser
from neural_insights.components.graph.graph import Graph
from neural_insights.components.graph.node import GroupNode, Node


def _create_simple_node(name: str, groups: List[str]) -> Node:
    """Create Node."""
    return Node(
        id=name,
        label=name,
        groups=groups,
    )


def _create_group_node(group_name: str) -> GroupNode:
    """Create Node."""
    node_name = f"node_group_{group_name}"
    return GroupNode(
        id=node_name,
        group_name=group_name,
    )


class TestCollapser(unittest.TestCase):
    """Test Collapser class."""

    def test_collapsing_with_all_groups_expanded(self) -> None:
        """Test that no collapsing will happen when all groups are expanded."""
        source = self._get_source_graph()

        collapser = Collapser(["a", "a:b", "a:b:c"])
        collapsed = collapser.collapse(graph=source)

        self.assertEqual(source, collapsed)

    def test_collapsing_all_groups(self) -> None:
        """Test that no collapsing will happen when all groups are expanded."""
        source = self._get_source_graph()

        expected = Graph()
        expected.add_node(_create_simple_node("input", []))
        expected.add_node(_create_group_node("a"))
        expected.add_node(_create_simple_node("e", ["e"]))
        expected.add_node(_create_simple_node("output", []))
        expected.add_edge("input", "node_group_a")
        expected.add_edge("node_group_a", "output")
        expected.add_edge("input", "e")
        expected.add_edge("e", "output")

        collapser = Collapser([])
        collapsed = collapser.collapse(graph=source)

        self.assertEqual(expected, collapsed)

    def test_expanding_all_groups(self) -> None:
        """Test that no collapsing will happen when all groups are expanded."""
        source = self._get_source_graph()

        collapser = Collapser(["node_group_a", "node_group_a:b", "node_group_a:b:c"])
        collapsed = collapser.collapse(graph=source)

        self.assertEqual(source, collapsed)

    def test_collapsing_external_group(self) -> None:
        """Test that no collapsing will happen when all groups are expanded."""
        source = self._get_source_graph()

        expected = Graph()
        expected.add_node(_create_simple_node("input", []))
        expected.add_node(_create_simple_node("a:b", ["a"]))
        expected.add_node(_create_group_node("a:b"))
        expected.add_node(_create_simple_node("e", ["e"]))
        expected.add_node(_create_simple_node("output", []))
        expected.add_edge("input", "a:b")
        expected.add_edge("a:b", "node_group_a:b")
        expected.add_edge("node_group_a:b", "output")
        expected.add_edge("input", "e")
        expected.add_edge("e", "output")

        collapser = Collapser(["node_group_a"])
        collapsed = collapser.collapse(graph=source)

        self.assertEqual(expected, collapsed)

    def test_collapsing_internal_group(self) -> None:
        """Test that no collapsing will happen when all groups are expanded."""
        source = self._get_source_graph()

        expected = Graph()
        expected.add_node(_create_simple_node("input", []))
        expected.add_node(_create_group_node("a"))
        expected.add_node(_create_simple_node("e", ["e"]))
        expected.add_node(_create_simple_node("output", []))
        expected.add_edge("input", "node_group_a")
        expected.add_edge("node_group_a", "output")
        expected.add_edge("input", "e")
        expected.add_edge("e", "output")

        collapser = Collapser(["node_group_a:b"])
        collapsed = collapser.collapse(graph=source)

        self.assertEqual(expected, collapsed)

    def _get_source_graph(self) -> Graph:
        """Return a graph with a well-known contents."""
        graph = Graph()
        graph.add_node(_create_simple_node("input", []))
        graph.add_node(_create_simple_node("a:b", ["a"]))
        graph.add_node(_create_simple_node("a:b:c1", ["a", "a:b"]))
        graph.add_node(_create_simple_node("a:b:c2", ["a", "a:b"]))
        graph.add_node(_create_simple_node("a:b:c:d", ["a", "a:b", "a:b:c"]))
        graph.add_node(_create_simple_node("e", ["e"]))
        graph.add_node(_create_simple_node("output", []))

        graph.add_edge("input", "a:b")
        graph.add_edge("a:b", "a:b:c1")
        graph.add_edge("a:b", "a:b:c2")
        graph.add_edge("a:b:c1", "a:b:c:d")
        graph.add_edge("a:b:c2", "a:b:c:d")
        graph.add_edge("a:b:c:d", "output")
        graph.add_edge("input", "e")
        graph.add_edge("e", "output")

        return graph


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""WSGI Web Server."""
import sys

import gevent.monkey

already_patched = gevent.monkey.is_module_patched("threading")
if not already_patched:
    can_patch_ssl = "ssl" not in sys.modules
    gevent.monkey.patch_all(ssl=can_patch_ssl)

from neural_insights.utils.exceptions import NotFoundException  # noqa: E402
from neural_insights.utils.logger import change_log_level, log  # noqa: E402
from neural_insights.web.configuration import Configuration  # noqa: E402
from neural_insights.web.server import run_server  # noqa: E402


def main() -> None:
    """Get parameters and initialize server."""
    try:
        configuration = Configuration()
    except NotFoundException as e:
        print(str(e))
        sys.exit(1)

    change_log_level(configuration.log_level)

    log.info("Neural Insights Server started.\n")

    if configuration.allow_insecure_connections:
        log.warning(
            "Running in insecure mode.\n" "Everyone in your network may attempt to access this server.\n",
        )

    log.info(f"Open address {configuration.get_url()}")
    configuration.dump_token_to_file()

    run_server(configuration)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch model class."""
import re
from typing import List

from neural_insights.components.model.model import Model
from neural_insights.utils.consts import Frameworks
from neural_insights.utils.utils import check_module, get_file_extension


class PyTorchScriptModel(Model):
    """PyTorch Script Model class."""

    def __init__(self, path: str) -> None:
        """Initialize object."""
        super().__init__(path)

    @staticmethod
    def _has_any_name_parts(nodes: set, name_parts: List[str]) -> bool:
        """Check if there is at least one node for name_parts."""
        matching_names = []
        for node in nodes:
            for partial_name in name_parts:
                search = re.match(partial_name, node)
                if search:
                    matching_names.append(node)
        return bool(matching_names)

    @staticmethod
    def get_framework_name() -> str:
        """Get the name of framework."""
        return Frameworks.PT.value

    @staticmethod
    def supports_path(path: str) -> bool:
        """Check if given path is of supported model."""
        return "py" == get_file_extension(path)

    def guard_requirements_installed(self) -> None:
        """Ensure all requirements are installed."""
        check_module("torch")



# -*- coding: utf-8 -*-
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Model repository."""

from typing import List, Type

from neural_insights.components.model.model import Model
from neural_insights.components.model.onnxrt.model import OnnxrtModel
from neural_insights.components.model.pytorch.model import PyTorchModel
from neural_insights.components.model.pytorch.pytorch_script.model import PyTorchScriptModel
from neural_insights.components.model.tensorflow.frozen_pb import FrozenPbModel
from neural_insights.components.model.tensorflow.keras import KerasModel
from neural_insights.components.model.tensorflow.meta_graph import MetaGraphModel
from neural_insights.components.model.tensorflow.saved_model import SavedModelModel
from neural_insights.utils.exceptions import NotFoundException
from neural_insights.utils.logger import log


class ModelRepository:
    """Model repository."""

    def __init__(self) -> None:
        """Initialize class."""
        self.model_types: List[Type[Model]] = [
            FrozenPbModel,
            KerasModel,
            MetaGraphModel,
            OnnxrtModel,
            SavedModelModel,
            PyTorchScriptModel,
            PyTorchModel,
        ]

    def get_model(self, path: str) -> Model:
        """Get Model for given path."""
        for model_type in self.model_types:
            supports_path = model_type.supports_path(path)
            log.debug(f"{model_type.__name__}: {supports_path}")
            if model_type.supports_path(path):
                return model_type(path)

        raise NotFoundException(f"Unable to find Model for path {path}")

    def get_frameworks(self) -> List[str]:
        """Get list of supported frameworks."""
        unique_frameworks = {model.get_framework_name() for model in self.model_types}
        return sorted(unique_frameworks)

    @staticmethod
    def get_framework_from_path(model_path: str) -> str:
        """Get framework name from model extension.

        :param model_path: Path to model.
        """
        model_repository = ModelRepository()
        try:
            model = model_repository.get_model(model_path)
            return model.get_framework_name()
        except NotFoundException:
            raise NotFoundException("Framework was not recognized.")

    @staticmethod
    def get_supported_frameworks() -> List[str]:
        """Get list of supported frameworks."""
        return ModelRepository().get_frameworks()

    @staticmethod
    def is_model_path(path: str) -> bool:
        """Check if provided path is for model."""
        try:
            ModelRepository().get_model(path)
            return True
        except NotFoundException:
            return False



# -*- coding: utf-8 -*-
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Abstract Tensorflow model class."""
import os.path
from typing import Any, List, Optional

from neural_compressor.experimental.common.model import Model as NCModel
from neural_compressor.model.model import TensorflowBaseModel
from neural_compressor.utils.logger import Logger
from neural_insights.components.graph.graph import Graph
from neural_insights.components.graph.reader.tensorflow_reader import TensorflowReader
from neural_insights.components.model.model import Model
from neural_insights.components.model.shape import Shape
from neural_insights.components.model.tensorflow.utils import get_input_shape
from neural_insights.utils.consts import Domains, Frameworks
from neural_insights.utils.logger import log
from neural_insights.utils.utils import check_module


class TensorflowModel(Model):
    """Abstract TensorFlow Model class."""

    def __init__(self, path: str) -> None:
        """Initialize object."""
        super().__init__(path)
        self._nc_model_instance: Optional[TensorflowBaseModel] = None

    def get_input_nodes(self) -> Optional[List[Any]]:
        """Get model input nodes."""
        self.guard_requirements_installed()

        # pylint: disable=maybe-no-member
        return self.nc_model_instance.input_node_names

    def get_output_nodes(self) -> Optional[List[Any]]:
        """Get model output nodes."""
        self.guard_requirements_installed()

        # pylint: disable=maybe-no-member
        return self.nc_model_instance.output_node_names + ["custom"]

    def get_model_graph(self) -> Graph:
        """Get model Graph."""
        graph_reader = TensorflowReader(self)
        return graph_reader.read()

    @property
    def nc_model_instance(self) -> TensorflowBaseModel:
        """Get Neural Compressor Model instance."""
        self._ensure_nc_model_instance()
        return self._nc_model_instance

    def _ensure_nc_model_instance(self) -> None:
        """Create Neural Compressor Model instance if needed."""
        if self._nc_model_instance is not None:
            return
        model_name = os.path.splitext(os.path.basename(self.path))[0]
        Logger().get_logger().setLevel(log.level)
        self._nc_model_instance = NCModel(self.path)
        self._nc_model_instance.name = model_name

    @property
    def input_shape(self) -> Shape:
        """Try to detect data shape."""
        try:
            domain = Domains(self.domain.domain)
            default_shapes = {
                Domains.IMAGE_RECOGNITION: 224,
                Domains.OBJECT_DETECTION: 300,
            }
            default_shape = default_shapes.get(domain, None)
        except ValueError:
            log.debug(f'Could not detect "{self.domain.domain}" domain.')
            default_shape = None

        if default_shape is None:
            return Shape(trusted=True)

        input_shape = [default_shape, default_shape, 3]
        is_trusted = False
        try:
            # pylint: disable=maybe-no-member
            input_shape = get_input_shape(self.nc_model_instance.graph_def, default_shape)
            is_trusted = True
        except Exception:
            log.debug(f"Could not get input shape from model. Using default shape: {input_shape}")

        return Shape(shape=", ".join(map(str, input_shape)), trusted=is_trusted)

    @property
    def shape_elements_order(self) -> List[str]:
        """Get order of input shape channels."""
        return ["height", "width", "channels"]

    @staticmethod
    def get_framework_name() -> str:
        """Get the name of framework."""
        return Frameworks.TF.value

    @property
    def supports_pruning(self) -> bool:
        """Check if pruning is supported for the model."""
        try:
            self.nc_model_instance.get_all_weight_names()
            return True
        except AttributeError:
            return False

    def guard_requirements_installed(self) -> None:
        """Ensure all requirements are installed."""
        check_module("tensorflow")
